"""
SiliconLife Eyeflow
Class for log batch of extracted images from detection

Author: Alex Sobral de Freitas
"""

import os

import json
import datetime
import pytz
import random
import cv2
from bson import ObjectId

from eyeflow_sdk.file_access import FileAccess
import eyeflow_sdk.img_utils as img_utils
from eyeflow_sdk.log_obj import log
#----------------------------------------------------------------------------------------------------------------------------------

MAX_EXTRACT_FILES = 800
THUMB_SIZE = 128

def clear_log(extract_path, max_files=MAX_EXTRACT_FILES):
    files_list = os.listdir(extract_path)
    if len(files_list) > max_files:
        date_list = [(filename, datetime.datetime.fromtimestamp(os.path.getmtime(os.path.join(extract_path, filename)))) for filename in files_list]
        exclude_list = sorted(date_list, key=lambda x: x[1])[:len(files_list) - max_files]
        for filename, _ in exclude_list:
            try:
                os.remove(os.path.join(extract_path, filename))
            except:
                pass
#----------------------------------------------------------------------------------------------------------------------------------


class VideoLog(object):
    def __init__(self, dataset_id, vlog_size, max_output_size=1920):
        self._vlog_size = vlog_size
        file_ac = FileAccess(storage="extract", resource_id=dataset_id)
        self._dest_path = file_ac.get_local_folder()
        self._dataset_id = dataset_id
        if isinstance(max_output_size, list) or isinstance(max_output_size, tuple):
            self._max_output_size = max(max_output_size)
        elif isinstance(max_output_size, int):
            self._max_output_size = max_output_size

        self._last_log = datetime.datetime(2000, 1, 1)


    def log_batch(self, image_batch, output_batch, annotations):
        for idx, image in enumerate(image_batch):
            if random.random() < float(self._vlog_size):
                obj_id = str(ObjectId())

                filename = obj_id + '.jpg'
                file_thumb = filename[:-4] + "_thumb.jpg"
                img = image["input_image"]
                scale = None
                if max(img.shape) > self._max_output_size:
                    img, scale = img_utils.resize_image_scale(img, self._max_output_size)

                cv2.imwrite(os.path.join(self._dest_path, filename), img)
                file_stat_img = os.stat(os.path.join(self._dest_path, filename))

                if max(img.shape) > THUMB_SIZE:
                    img_thumb, _ = img_utils.resize_image_scale(img, THUMB_SIZE)
                    cv2.imwrite(os.path.join(self._dest_path, file_thumb), img_thumb)
                else:
                    cv2.imwrite(os.path.join(self._dest_path, file_thumb), img)
                file_stat_thumb = os.stat(os.path.join(self._dest_path, file_thumb))

                if scale is not None and "instances" in annotations[idx]:
                    for idx_ann, ann in enumerate(annotations[idx]["instances"]):
                        if "bbox" in ann:
                            annotations[idx]["instances"][idx_ann]["bbox"]["x_min"] *= scale
                            annotations[idx]["instances"][idx_ann]["bbox"]["y_min"] *= scale
                            annotations[idx]["instances"][idx_ann]["bbox"]["x_max"] *= scale
                            annotations[idx]["instances"][idx_ann]["bbox"]["y_max"] *= scale

                img_data = {
                    "_id": obj_id,
                    "date": pytz.utc.localize(datetime.datetime.now()),
                    "img_height": img.shape[0],
                    "img_width": img.shape[1],
                    "file_size": file_stat_img.st_size,
                    "thumb_size": file_stat_thumb.st_size,
                    "annotations": annotations[idx]
                }

                if 'frame_time' in image["frame_data"]:
                    img_data['frame_time'] = image["frame_data"]['frame_time']

                if 'video_data' in image["frame_data"]:
                    img_data['video_data'] = image["frame_data"]['video_data']

                with open(os.path.join(self._dest_path, obj_id + '_data.json'), 'w', newline='', encoding='utf8') as file_p:
                    json.dump(img_data, file_p, ensure_ascii=False, default=str)

                if (datetime.datetime.now() - self._last_log) > datetime.timedelta(minutes=1):
                    clear_log(self._dest_path)
                    self._last_log = datetime.datetime.now()
#----------------------------------------------------------------------------------------------------------------------------------



import numpy as np
import copy
# ---------------------------------------------------------------------------------------------------------------------------------

def intersection_over_union(p_boxA, p_boxB):
    # determine the (x, y)-coordinates of the intersection rectangle
    xA = max(p_boxA["x_min"], p_boxB["x_min"])
    yA = max(p_boxA["y_min"], p_boxB["y_min"])
    xB = min(p_boxA["x_max"], p_boxB["x_max"])
    yB = min(p_boxA["y_max"], p_boxB["y_max"])

    # compute the area of intersection rectangle
    interArea = max(0, xB - xA) * max(0, yB - yA)

    # compute the area of both the prediction and ground-truth rectangles
    boxAArea = (p_boxA["x_max"] - p_boxA["x_min"] + 1) * (p_boxA["y_max"] - p_boxA["y_min"] + 1)
    boxBArea = (p_boxB["x_max"] - p_boxB["x_min"] + 1) * (p_boxB["y_max"] - p_boxB["y_min"] + 1)

    # compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the interesection area
    if (boxAArea + boxBArea - interArea) > 0:
        iou = interArea / float(boxAArea + boxBArea - interArea)
    else:
        iou = 0

    # return the intersection over union value
    return iou
# ---------------------------------------------------------------------------------------------------------------------------------


class mAP():
    def __init__(self, min_iou=0.75):
        self._map_list = []
        self._total_gt = 0
        self._min_iou = min_iou

    def add_detection(self, y_pred, y_true):
        gtp = copy.deepcopy(y_true)
        preds = copy.deepcopy(y_pred)
        self._total_gt += len(gtp)
        tpd = {}
        fpd = {}
        fnd = {}
        for idx_gtp, tr in enumerate(gtp):
            found_tr = False
            for idx_pr, pr in enumerate(preds):
                iou = intersection_over_union(pr["bbox"], tr["bbox"])
                if iou > self._min_iou:
                    found_tr = True
                    if pr["class"] == tr["class"]:
                        if idx_gtp not in tpd.keys():
                            tpd[idx_gtp] = (tr, pr)
                            del(preds[idx_pr])
                        elif idx_gtp not in fpd.keys():
                            fpd[idx_gtp] = [tpd[idx_gtp], (tr, pr)]
                            tpd[idx_gtp] = None
                            del(preds[idx_pr])
                        else:
                            fpd[idx_gtp].append((tr, pr))
                            del(preds[idx_pr])
                    else:
                        if idx_gtp not in fnd.keys():
                            fnd[idx_gtp] = [(tr, pr)]
                            del(preds[idx_pr])
                        else:
                            fnd[idx_gtp].append((tr, pr))
                            del(preds[idx_pr])

            if found_tr:
                del(gtp[idx_gtp])

        fpl = [pr for pr in preds]

        TP = len(tpd)
        FP = len(fpd) + len(fpl)
        FN = len(fnd) + len(gtp)

        self._map_list.append((TP, FP, FN))

    def add_fn(self, y_true):
        self._map_list.append((0, 0, len(y_true)))

    def get_map(self):
        tp_acc = 0
        fp_acc = 0
        fn_acc = 0
        precision = []
        recall = []
        for tp, fp, fn in self._map_list:
            tp_acc += tp
            fp_acc += fp
            fn_acc += fn
            if (tp_acc + fp_acc) > 0:
                precision.append(tp_acc / (tp_acc + fp_acc))
            else:
                precision.append(0)

            if self._total_gt > 0:
                recall.append(tp_acc / self._total_gt)
            else:
                recall.append(0)

        precision = np.array(precision)
        recall = np.array(recall)
        p_interp = []
        for r in range(11):
            indices = np.nonzero((recall >= r/10) & (recall < (r + 1)/10))
            if indices[0].shape[0] > 0:
                last_indices = indices
            p_interp.append(np.max(precision[last_indices]))

        return np.mean(np.array(p_interp))
# ---------------------------------------------------------------------------------------------------------------------------------
